@@ -18,7 +18,7 @@
 std::string getVersion()
 {
 	char result[32];
-	sprintf(result, "%d.%d", MESHOPTIMIZER_VERSION / 1000, (MESHOPTIMIZER_VERSION % 1000) / 10);
+	snprintf(result, sizeof(result), "%d.%d", MESHOPTIMIZER_VERSION / 1000, (MESHOPTIMIZER_VERSION % 1000) / 10);
 	return result;
 }
 
@@ -250,12 +250,12 @@ static bool canTransformMesh(const Mesh& mesh)
 	return true;
 }
 
-static void process(cgltf_data* data, const char* input_path, const char* output_path, const char* report_path, std::vector<Mesh>& meshes, std::vector<Animation>& animations, const std::string& extras, const Settings& settings, std::string& json, std::string& bin, std::string& fallback, size_t& fallback_size)
+static void process(cgltf_data* data, const char* input_path, const char* output_path, const char* report_path, std::vector<Mesh>& meshes, std::vector<Animation>& animations, const Settings& settings, std::string& json, std::string& bin, std::string& fallback, size_t& fallback_size)
 {
 	if (settings.verbose)
 	{
-		printf("input: %d nodes, %d meshes (%d primitives), %d materials, %d skins, %d animations\n",
-		    int(data->nodes_count), int(data->meshes_count), int(meshes.size()), int(data->materials_count), int(data->skins_count), int(animations.size()));
+		printf("input: %d nodes, %d meshes (%d primitives), %d materials, %d skins, %d animations, %d images\n",
+		    int(data->nodes_count), int(data->meshes_count), int(meshes.size()), int(data->materials_count), int(data->skins_count), int(animations.size()), int(data->images_count));
 		printMeshStats(meshes, "input");
 	}
 
@@ -432,6 +432,8 @@ static void process(cgltf_data* data, const char* input_path, const char* output
 	bool ext_specular = false;
 	bool ext_sheen = false;
 	bool ext_volume = false;
+	bool ext_emissive_strength = false;
+	bool ext_iridescence = false;
 	bool ext_unlit = false;
 	bool ext_instancing = false;
 	bool ext_texture_transform = false;
@@ -451,21 +453,36 @@ static void process(cgltf_data* data, const char* input_path, const char* output
 		append(json_samplers, "}");
 	}
 
+	std::vector<std::string> encoded_images;
+
+#ifdef WITH_BASISU
+	if (data->images_count && settings.texture_ktx2)
+	{
+		encoded_images.resize(data->images_count);
+
+		encodeImages(encoded_images.data(), data, images, input_path, settings);
+	}
+#endif
+
 	for (size_t i = 0; i < data->images_count; ++i)
 	{
 		const cgltf_image& image = data->images[i];
 
-		if (settings.verbose == 1 && settings.texture_ktx2)
-		{
-			const char* uri = image.uri;
-			bool embedded = !uri || strncmp(uri, "data:", 5) == 0;
-
-			printf("image %d (%s) is being encoded with %s\n", int(i), embedded ? "embedded" : uri, settings.texture_toktx ? "toktx" : "basisu");
-		}
-
 		comma(json_images);
 		append(json_images, "{");
-		writeImage(json_images, views, image, images[i], i, input_path, output_path, settings);
+		if (encoded_images.size() && !encoded_images[i].empty())
+		{
+			if (encoded_images[i].compare(0, 5, "error") == 0)
+				fprintf(stderr, "Warning: unable to encode image %d (%s), skipping (%s)\n", int(i), image.uri ? image.uri : "?", encoded_images[i].c_str());
+			else
+				writeEncodedImage(json_images, views, image, encoded_images[i], images[i], output_path, settings);
+
+			encoded_images[i] = std::string(); // reclaim memory early
+		}
+		else
+		{
+			writeImage(json_images, views, image, images[i], i, input_path, settings);
+		}
 		append(json_images, "}");
 	}
 
@@ -475,7 +492,7 @@ static void process(cgltf_data* data, const char* input_path, const char* output
 
 		comma(json_textures);
 		append(json_textures, "{");
-		writeTexture(json_textures, texture, data, settings);
+		writeTexture(json_textures, texture, texture.image ? &images[texture.image - data->images] : NULL, data, settings);
 		append(json_textures, "}");
 	}
 
@@ -490,9 +507,9 @@ static void process(cgltf_data* data, const char* input_path, const char* output
 
 		comma(json_materials);
 		append(json_materials, "{");
-		writeMaterial(json_materials, data, material, settings.quantize ? &qp : NULL, settings.quantize ? &qt_materials[i] : NULL);
+		writeMaterial(json_materials, data, material, settings.quantize && !settings.pos_float ? &qp : NULL, settings.quantize ? &qt_materials[i] : NULL);
 		if (settings.keep_extras)
-			writeExtras(json_materials, extras, material.extras);
+			writeExtras(json_materials, material.extras);
 		append(json_materials, "}");
 
 		mi.remap = int(material_offset);
@@ -505,6 +522,8 @@ static void process(cgltf_data* data, const char* input_path, const char* output
 		ext_specular = ext_specular || material.has_specular;
 		ext_sheen = ext_sheen || material.has_sheen;
 		ext_volume = ext_volume || material.has_volume;
+		ext_emissive_strength = ext_emissive_strength || material.has_emissive_strength;
+		ext_iridescence = ext_iridescence || material.has_iridescence;
 		ext_unlit = ext_unlit || material.unlit;
 		ext_texture_transform = ext_texture_transform || mi.usesTextureTransform;
 	}
@@ -634,13 +653,23 @@ static void process(cgltf_data* data, const char* input_path, const char* output
 			for (size_t j = 0; j < mesh.nodes.size(); ++j)
 			{
 				NodeInfo& ni = nodes[mesh.nodes[j] - data->nodes];
-
 				assert(ni.keep);
-				ni.meshes.push_back(node_offset);
 
-				writeMeshNode(json_nodes, mesh_offset, mesh.nodes[j], mesh.skin, data, settings.quantize ? &qp : NULL);
+				// if we don't use position quantization, prefer attaching the mesh to its node directly
+				if (!ni.has_mesh && (!settings.quantize || settings.pos_float))
+				{
+					ni.has_mesh = true;
+					ni.mesh_index = mesh_offset;
+					ni.mesh_skin = mesh.skin;
+				}
+				else
+				{
+					ni.mesh_nodes.push_back(node_offset);
 
-				node_offset++;
+					writeMeshNode(json_nodes, mesh_offset, mesh.nodes[j], mesh.skin, data, settings.quantize && !settings.pos_float ? &qp : NULL);
+
+					node_offset++;
+				}
 			}
 		}
 		else if (mesh.instances.size())
@@ -662,7 +691,7 @@ static void process(cgltf_data* data, const char* input_path, const char* output
 			comma(json_roots[mesh.scene]);
 			append(json_roots[mesh.scene], node_offset);
 
-			writeMeshNode(json_nodes, mesh_offset, NULL, mesh.skin, data, settings.quantize ? &qp : NULL);
+			writeMeshNode(json_nodes, mesh_offset, NULL, mesh.skin, data, settings.quantize && !settings.pos_float ? &qp : NULL);
 
 			node_offset++;
 		}
@@ -689,7 +718,7 @@ static void process(cgltf_data* data, const char* input_path, const char* output
 		append(json_nodes, "{");
 		writeNode(json_nodes, node, nodes, data);
 		if (settings.keep_extras)
-			writeExtras(json_nodes, extras, node.extras);
+			writeExtras(json_nodes, node.extras);
 		append(json_nodes, "}");
 	}
 
@@ -767,7 +796,7 @@ static void process(cgltf_data* data, const char* input_path, const char* output
 	append(json, "\"version\":\"2.0\",\"generator\":\"gltfpack ");
 	append(json, getVersion());
 	append(json, "\"");
-	writeExtras(json, extras, data->asset.extras);
+	writeExtras(json, data->asset.extras);
 	append(json, "}");
 
 	const ExtensionInfo extensions[] = {
@@ -781,6 +810,8 @@ static void process(cgltf_data* data, const char* input_path, const char* output
 	    {"KHR_materials_specular", ext_specular, false},
 	    {"KHR_materials_sheen", ext_sheen, false},
 	    {"KHR_materials_volume", ext_volume, false},
+	    {"KHR_materials_emissive_strength", ext_emissive_strength, false},
+	    {"KHR_materials_iridescence", ext_iridescence, false},
 	    {"KHR_materials_unlit", ext_unlit, false},
 	    {"KHR_materials_variants", data->variants_count > 0, false},
 	    {"KHR_lights_punctual", data->lights_count > 0, false},
@@ -917,7 +948,6 @@ int gltfpack(const char* input, const char* output, const char* report, Settings
 	cgltf_data* data = 0;
 	std::vector<Mesh> meshes;
 	std::vector<Animation> animations;
-	std::string extras;
 
 	std::string iext = getExtension(input);
 	std::string oext = output ? getExtension(output) : "";
@@ -925,7 +955,7 @@ int gltfpack(const char* input, const char* output, const char* report, Settings
 	if (iext == ".gltf" || iext == ".glb")
 	{
 		const char* error = 0;
-		data = parseGltf(input, meshes, animations, extras, &error);
+		data = parseGltf(input, meshes, animations, &error);
 
 		if (error)
 		{
@@ -950,31 +980,16 @@ int gltfpack(const char* input, const char* output, const char* report, Settings
 		return 2;
 	}
 
+#ifndef WITH_BASISU
 	if (data->images_count && settings.texture_ktx2)
 	{
-		if (checkKtx(settings.verbose > 1))
-		{
-			settings.texture_toktx = true;
-		}
-		else if (!checkBasis(settings.verbose > 1))
-		{
-			fprintf(stderr, "Error: toktx is not present in PATH or TOKTX_PATH is not set\n");
-			fprintf(stderr, "Note: toktx must be installed manually from https://github.com/KhronosGroup/KTX-Software/releases\n");
-			return 3;
-		}
-
-		if (settings.texture_scale < 1 && !settings.texture_toktx)
-		{
-			fprintf(stderr, "Error: -ts option is only supported by toktx\n");
-			return 3;
-		}
-
-		if (settings.texture_pow2 && !settings.texture_toktx)
-		{
-			fprintf(stderr, "Error: -tp option is only supported by toktx\n");
-			return 3;
-		}
+		fprintf(stderr, "Error: gltfpack was built without BasisU support, texture compression is not available\n");
+#ifdef __wasi__
+		fprintf(stderr, "Note: node.js builds do not support BasisU due to lack of platform features; download a native build from https://github.com/zeux/meshoptimizer/releases\n");
+#endif
+		return 3;
 	}
+#endif
 
 	if (oext == ".glb")
 	{
@@ -983,7 +998,7 @@ int gltfpack(const char* input, const char* output, const char* report, Settings
 
 	std::string json, bin, fallback;
 	size_t fallback_size = 0;
-	process(data, input, output, report, meshes, animations, extras, settings, json, bin, fallback, fallback_size);
+	process(data, input, output, report, meshes, animations, settings, json, bin, fallback, fallback_size);
 
 	cgltf_free(data);
 
@@ -1181,6 +1196,20 @@ int main(int argc, char** argv)
 		{
 			settings.col_bits = clamp(atoi(argv[++i]), 1, 16);
 		}
+		else if (strcmp(arg, "-vpi") == 0)
+		{
+			settings.pos_float = false;
+			settings.pos_normalized = false;
+		}
+		else if (strcmp(arg, "-vpn") == 0)
+		{
+			settings.pos_float = false;
+			settings.pos_normalized = true;
+		}
+		else if (strcmp(arg, "-vpf") == 0)
+		{
+			settings.pos_float = true;
+		}
 		else if (strcmp(arg, "-at") == 0 && i + 1 < argc && isdigit(argv[i + 1][0]))
 		{
 			settings.trn_bits = clamp(atoi(argv[++i]), 1, 24);
@@ -1229,6 +1258,10 @@ int main(int argc, char** argv)
 		{
 			settings.simplify_aggressive = true;
 		}
+		else if (strcmp(arg, "-slb") == 0)
+		{
+			settings.simplify_lock_borders = true;
+		}
 #ifndef NDEBUG
 		else if (strcmp(arg, "-sd") == 0 && i + 1 < argc && isdigit(argv[i + 1][0]))
 		{
@@ -1249,11 +1282,19 @@ int main(int argc, char** argv)
 
 			for (int kind = 0; kind < TextureKind__Count; ++kind)
 				if (mask & (1 << kind))
-					settings.texture_uastc[kind] = true;
+					settings.texture_mode[kind] = TextureMode_UASTC;
 		}
 		else if (strcmp(arg, "-tc") == 0)
 		{
 			settings.texture_ktx2 = true;
+
+			unsigned int mask = ~0u;
+			if (i + 1 < argc && isalpha(argv[i + 1][0]))
+				mask = textureMask(argv[++i]);
+
+			for (int kind = 0; kind < TextureKind__Count; ++kind)
+				if (mask & (1 << kind))
+					settings.texture_mode[kind] = TextureMode_ETC1S;
 		}
 		else if (strcmp(arg, "-tq") == 0 && i + 1 < argc && isdigit(argv[i + 1][0]))
 		{
@@ -1274,6 +1315,10 @@ int main(int argc, char** argv)
 		{
 			settings.texture_scale = clamp(float(atof(argv[++i])), 0.f, 1.f);
 		}
+		else if (strcmp(arg, "-tl") == 0 && i + 1 < argc && isdigit(argv[i + 1][0]))
+		{
+			settings.texture_limit = atoi(argv[++i]);
+		}
 		else if (strcmp(arg, "-tp") == 0)
 		{
 			settings.texture_pow2 = true;
@@ -1282,12 +1327,13 @@ int main(int argc, char** argv)
 		{
 			settings.texture_flipy = true;
 		}
-		else if (strcmp(arg, "-te") == 0)
+		else if (strcmp(arg, "-tj") == 0 && i + 1 < argc && isdigit(argv[i + 1][0]))
 		{
-			fprintf(stderr, "Warning: -te is deprecated and will be removed in the future; gltfpack now automatically embeds textures into GLB files\n");
+			settings.texture_jobs = clamp(atoi(argv[++i]), 0, 128);
 		}
 		else if (strcmp(arg, "-noq") == 0)
 		{
+			// TODO: Warn if -noq is used and suggest -vpf instead; use -noqq to silence
 			settings.quantize = false;
 		}
 		else if (strcmp(arg, "-i") == 0 && i + 1 < argc && !input)
@@ -1380,24 +1426,32 @@ int main(int argc, char** argv)
 			fprintf(stderr, "\t-o file: output file path, .gltf/.glb\n");
 			fprintf(stderr, "\t-c: produce compressed gltf/glb files (-cc for higher compression ratio)\n");
 			fprintf(stderr, "\nTextures:\n");
-			fprintf(stderr, "\t-tc: convert all textures to KTX2 with BasisU supercompression (using basisu/toktx executable)\n");
+			fprintf(stderr, "\t-tc: convert all textures to KTX2 with BasisU supercompression\n");
 			fprintf(stderr, "\t-tu: use UASTC when encoding textures (much higher quality and much larger size)\n");
 			fprintf(stderr, "\t-tq N: set texture encoding quality (default: 8; N should be between 1 and 10\n");
 			fprintf(stderr, "\t-ts R: scale texture dimensions by the ratio R (default: 1; R should be between 0 and 1)\n");
+			fprintf(stderr, "\t-tl N: limit texture dimensions to N pixels (default: 0 = no limit)\n");
 			fprintf(stderr, "\t-tp: resize textures to nearest power of 2 to conform to WebGL1 restrictions\n");
 			fprintf(stderr, "\t-tfy: flip textures along Y axis during BasisU supercompression\n");
+			fprintf(stderr, "\t-tj N: use N threads when compressing textures\n");
 			fprintf(stderr, "\tTexture classes:\n");
+			fprintf(stderr, "\t-tc C: use ETC1S when encoding textures of class C\n");
 			fprintf(stderr, "\t-tu C: use UASTC when encoding textures of class C\n");
 			fprintf(stderr, "\t-tq C N: set texture encoding quality for class C\n");
 			fprintf(stderr, "\t... where C is a comma-separated list (no spaces) with valid values color,normal,attrib\n");
 			fprintf(stderr, "\nSimplification:\n");
-			fprintf(stderr, "\t-si R: simplify meshes to achieve the ratio R (default: 1; R should be between 0 and 1)\n");
+			fprintf(stderr, "\t-si R: simplify meshes targeting triangle count ratio R (default: 1; R should be between 0 and 1)\n");
 			fprintf(stderr, "\t-sa: aggressively simplify to the target ratio disregarding quality\n");
+			fprintf(stderr, "\t-slb: lock border vertices during simplification to avoid gaps on connected meshes\n");
 			fprintf(stderr, "\nVertices:\n");
 			fprintf(stderr, "\t-vp N: use N-bit quantization for positions (default: 14; N should be between 1 and 16)\n");
 			fprintf(stderr, "\t-vt N: use N-bit quantization for texture coordinates (default: 12; N should be between 1 and 16)\n");
 			fprintf(stderr, "\t-vn N: use N-bit quantization for normals and tangents (default: 8; N should be between 1 and 16)\n");
 			fprintf(stderr, "\t-vc N: use N-bit quantization for colors (default: 8; N should be between 1 and 16)\n");
+			fprintf(stderr, "\nVertex positions:\n");
+			fprintf(stderr, "\t-vpi: use integer attributes for positions (default)\n");
+			fprintf(stderr, "\t-vpn: use normalized attributes for positions\n");
+			fprintf(stderr, "\t-vpf: use floating point attributes for positions\n");
 			fprintf(stderr, "\nAnimations:\n");
 			fprintf(stderr, "\t-at N: use N-bit quantization for translations (default: 16; N should be between 1 and 24)\n");
 			fprintf(stderr, "\t-ar N: use N-bit quantization for rotations (default: 12; N should be between 4 and 16)\n");
@@ -1423,14 +1477,26 @@ int main(int argc, char** argv)
 			fprintf(stderr, "\t-i file: input file to process, .obj/.gltf/.glb\n");
 			fprintf(stderr, "\t-o file: output file path, .gltf/.glb\n");
 			fprintf(stderr, "\t-c: produce compressed gltf/glb files (-cc for higher compression ratio)\n");
-			fprintf(stderr, "\t-tc: convert all textures to KTX2 with BasisU supercompression (using basisu/toktx executable)\n");
-			fprintf(stderr, "\t-si R: simplify meshes to achieve the ratio R (default: 1; R should be between 0 and 1)\n");
+			fprintf(stderr, "\t-tc: convert all textures to KTX2 with BasisU supercompression\n");
+			fprintf(stderr, "\t-si R: simplify meshes targeting triangle count ratio R (default: 1; R should be between 0 and 1)\n");
 			fprintf(stderr, "\nRun gltfpack -h to display a full list of options\n");
 		}
 
 		return 1;
 	}
 
+	if (settings.texture_limit && !settings.texture_ktx2)
+	{
+		fprintf(stderr, "Option -tl is only supported when -tc is set as well\n");
+		return 1;
+	}
+
+	if (settings.texture_pow2 && (settings.texture_limit & (settings.texture_limit - 1)) != 0)
+	{
+		fprintf(stderr, "Option -tp requires the limit specified via -tl to be a power of 2\n");
+		return 1;
+	}
+
 	if (settings.texture_scale < 1 && !settings.texture_ktx2)
 	{
 		fprintf(stderr, "Option -ts is only supported when -tc is set as well\n");
@@ -1449,6 +1515,18 @@ int main(int argc, char** argv)
 		return 1;
 	}
 
+	if (settings.fallback && settings.compressmore)
+	{
+		fprintf(stderr, "Option -cf can not be used together with -cc\n");
+		return 1;
+	}
+
+	if (settings.fallback && settings.pos_float)
+	{
+		fprintf(stderr, "Option -cf can not be used together with -vpf\n");
+		return 1;
+	}
+
 	return gltfpack(input, output, report, settings);
 }
 
